#include <iostream>
#include <vector>
#include <algorithm>
#include "lagrange.hh"
#include "luminosity.hh"
#include "math/integration.hh"
#include <gsl/gsl_math.h>
#include <gsl/gsl_sf.h>

using namespace std;


extern double resumm::LumRescaled(double x, void *p);

namespace lagrange{

int binomial(int n, int k)
{
  return gsl_sf_choose(n,k);
  double result=1.;
  for (int i = 1; i <= k; ++i)
  {
    result *= ((double)n+1.-(double)i)/((double)i);
  }
  //weird, it fails with n=15, k=3:
  //result = 455, (int) result = 454...
  //this should fix this issue 
  return (int) (result+0.1);
}

void lagrange_coeff(double * C, double * x, double * y, int n, int k)
{

  C[n-1] = 1.;

  for (int c = 1; c < n; ++c)
  {
    C[n-1-c] = 0;
    //strategy: place c selectors in v, create all permutations of these selectors
    //and save the corresponding set member if it is selected in in the
    //current permutation of v: taken from
    //http://stackoverflow.com/questions/9430568/generating-combinations-in-c
    vector<bool> v(n);
    fill(v.begin() + c, v.end(), true);

    //create a vector to allocate all possible combinations (bin(n,c))
    int **  combinations;
    int comb = binomial(n,c);
    combinations = new int * [comb];
    for (int i = 0; i < comb; ++i)
      combinations[i] = new int[c];

    //fill the vector with the combinations
    int counter = 0;
    int c_counter = 0;
    do {
      for (int i = 0; i < n; ++i) {
        if (!v[i] ) {
          combinations[counter][c_counter] = (i+1);
          c_counter ++;
        }
      }
      counter ++;
      c_counter = 0;
    } while (next_permutation(v.begin(), v.end()));

    //evaluate the numerator prod_(j \neq k) (x-x_j)
    for (int i = 0; i < comb; ++i)
    {
      double partial_res = 1;
      for (int j = 0; j < c; ++j)
      {
        partial_res *= x[combinations[i][j]-1];
        if(combinations[i][j]-1 == k)
          partial_res *= 0; // c \neq k
      }
      C[n-1-c] += pow(-1,c)*partial_res;
    }
  }
  
  //calculate denominator
  double denominator = 1;
  for (int i = 0; i < n; ++i)
  {
    if(i != k)
      denominator *= (x[k]-x[i]);  // i \neq k
  }
 
  //calculate lagrange coefficients
  for (int i = 0; i < n; ++i)
  {
    //cout << C[i] << "   ";
    C[i] *= y[k]/denominator;
    //cout << C[i] << endl;
  }

}

};



int Lagrange::approx(unsigned int n, double *x, double *coeff, void *pars) {

  using namespace lagrange;
  //
  double *f = new double[n];
  //
  for(unsigned int k=0; k<n; k++) {
    f[k] = fLum_log(x[k], pars);
  }

  //
  double ** lagrange_coefficients = new double * [n];
  for (int i = 0; i < n; ++i)
    lagrange_coefficients[i] = new double [n];

  //evaluate the coefficients
  for (int i = 0; i < n; ++i){
    lagrange_coeff(lagrange_coefficients[i],x,f,n,i);
    coeff[i] = 0;
  }

  for (int i = 0; i < n; ++i)
  {
    //cout<< lagrange_coefficients[2][i] << endl;
    for (int j = 0; j < n; ++j)
    {
      coeff[i]+= lagrange_coefficients[j][i];//check if i and j are ok
    }
  }


  return 1;
}





Lagrange::Lagrange(resumm::Resummation *r) {
  res = r;
  quality_check = false;
  return;
}
Lagrange::~Lagrange() {
  return;
}


double Lagrange::fLum_log(double x, void *p) {
  double z = exp(x);
  return resumm::LumRescaled(z,res) * pow(1-z,-gamma_lum) * pow(z,beta_lum);
}




/*
double recLum_func(double u, double *c, int N, double A, double B) {
  double res = 0.;
  for(int i=0; i<=N; i++) {
    res += c[i] * pow(A*u+B, i);
  }
  return res;
}
struct sQC {
  Lagrange *lagr;
  void *extra_pars;
  int N;
  double *c;
  double A, B;
};
double integrand_QC(double u, void *p) {
  sQC pars = *(sQC *) p;
  return fabs( pars.lagr->Lum_func(pars.type, u, pars.extra_pars)
	       //- pars.cheb->rec_Lum_func(pars.type, z, pars.c, pars.N);
	       - recLum_func(u, pars.c, pars.N, pars.A, pars.B)
	       );
}
*/




int Lagrange::ApproxLuminosity(double *coeff, unsigned int N, double beta, int gamma, double z_min) {
  using namespace lagrange;
  beta_lum = beta;
  gamma_lum = gamma;
  //
  //create an array log(z_min) -> 1
  double * x = new double[N+1];
  for (int i = 0; i<=N; ++i) {
    x[i] = log(z_min) + (log(0.99) - log(z_min))*i/N;
    //cout << i << "  " << exp(x[i]) << endl;
  }

  approx(N+1, x, coeff, NULL);
  for (int i = 0; i<=N; ++i) {
    coeff[i] *= pow(-1.,i) * gsl_sf_fact(i);
    //cout << i << "  " << coeff[i] << endl;
  }

  //
  /*
  if(quality_check) {
    double prec = 1.e-1;
    sQC pars = { this, NULL, N, coeff_tilde, 2./M, 1. };
    double qc = integration::gauss(integrand_QC, -M, 0., prec, NULL, &pars);
    cout << "Quality Check: this number should be as small as possible: " << qc << endl;
  }
  */
  return 1;
}


void Lagrange::EnableQualityCheck(bool enable) {
  quality_check = enable;
}
